import os
import pandas as pd


class Collector():
    def __init__(self, filename):
        self.filename = filename

    def collect(self, dataset_readers_list):

        if len(dataset_readers_list) == 0:
            return

        output = self._prepare_output(dataset_readers_list)
        output.to_hdf(self.filename, key='df', mode='w')

    def _prepare_output(self, dataset_readers_list):
        dataset_readers_list = [(d, [r.contents for r in readers]) for d, readers in dataset_readers_list if readers]
        if len(dataset_readers_list) == 0:
            return None

        return self._merge_dataframes(dataset_readers_list)

    def _merge_dataframes(self, dataset_readers_list):

        final_df = None
        all_dfs = []
        keys = []

        for dataset, readers in dataset_readers_list:
            dataset_df = readers[0]
            for df in readers[1:]:
                if df is None:
                    continue
                dataset_df = dataset_df.add(df, fill_value=0.)

            all_dfs.append(dataset_df)
            keys.append(dataset)

        final_df = pd.concat(all_dfs, keys=keys, names=['dataset'], sort=True)

        return final_df.reset_index()


class EventByEventDataframe(object):
    """
    Write out a pandas dataframe with event-level values
    """

    def __init__(self, name, out_dir, collections, mask=None, flatten=True):

        self.name = name
        self.out_dir = out_dir
        self.mask = mask
        self.collections = collections
        self.contents = None
        self.df = None
        self.flatten = flatten

    def event(self, chunk):
        variables = chunk.tree.pandas.df(self.collections, flatten=self.flatten)

        variables["hashed_filename"] = hash(chunk.config.inputPaths[0])

        self.df = variables

        if self.contents is None:
            self.contents = self.df
        else:
            self.contents = pd.concat([self.contents, self.df])

        return True

    def collector(self):

        outfilename = "df_" + self.name + ".hd5"
        outfilename = os.path.join(self.out_dir, outfilename)
        return Collector(outfilename)

    def merge(self, rhs):
        if rhs.contents is None or len(rhs.contents) == 0:
            return
        if self.contents is None:
            self.contents = rhs.contents
            return
        self.contents = pd.concat([self.contents, rhs.contents])



"""
Chop up those trees into nice little tables and dataframes
"""
from __future__ import print_function
import sys
from .help import help_stages
import fast_flow.v1 as fast_flow
import fast_curator
import logging
from .backends import get_backend
from .utils import mkdir_p
from .version import __version__
logging.getLogger(__name__).setLevel(logging.INFO)


def create_parser():
    from argparse import ArgumentParser, Action

    class StagesHelp(Action):
        def __call__(self, parser, namespace, values, option_string=None):
            full_output = option_string == "--help-stages-full"
            help_stages(values, full_output=full_output)
            sys.exit(0)

    parser = ArgumentParser(description=__doc__)
    parser.add_argument("dataset_cfg", type=str,
                        help="Dataset config to run over")
    parser.add_argument("sequence_cfg", type=str,
                        help="Config for how to process events")
    parser.add_argument("--outdir", default="output", type=str,
                        help="Where to save the results")
    parser.add_argument("--mode", default="multiprocessing", type=str,
                        help="Which mode to run in (multiprocessing, htcondor, sge)")
    parser.add_argument("--ncores", default=1, type=int,
                        help="Number of cores to run on")
    parser.add_argument("--nblocks-per-dataset", default=-1, type=int,
                        help="Number of blocks per dataset")
    parser.add_argument("--nblocks-per-sample", default=-1, type=int,
                        help="Number of blocks per sample")
    parser.add_argument("--blocksize", default=1000000, type=int,
                        help="Number of events per block")
    parser.add_argument("--quiet", default=False, action='store_true',
                        help="Keep progress report quiet")
    parser.add_argument("--profile", default=False, action='store_true',
                        help="Profile the code")
    parser.add_argument("--help-stages", nargs="?", default=None, action=StagesHelp,
                        metavar="stage-name-regex",
                        help="Print help specific to the available stages")
    parser.add_argument("--help-stages-full", action=StagesHelp, metavar="stage",
                        help="Print the full help specific to the available stages")
    parser.add_argument("-v", "--version", action="version", version='%(prog)s ' + __version__)

    return parser


def main(args=None):
    args = create_parser().parse_args(args)

    sequence = fast_flow.read_sequence_yaml(args.sequence_cfg, output_dir=args.outdir, backend="fast_carpenter")
    datasets = fast_curator.read.from_yaml(args.dataset_cfg)
    backend = get_backend(args.mode)

    mkdir_p(args.outdir)
    results, _ = backend.execute(sequence, datasets, args)

    print("Summary of results")
    print(results)
    print("Output written to directory '%s'" % args.outdir)
    return 0


if __name__ == "__main__":
    main()



"""Defines version of codebase
"""


def split_version(version):
    """Split a semantic version string into a version_info tuple
    """
    result = (version,)
    for div in ".-":
        result = [tok.split(div) for tok in result]
        result = sum(result, [])
    return tuple(result)


__version__ = '0.17.4'
version_info = split_version(__version__) # noqa



"""
Provides a common interface to select a backend

Each backend is wrapped in a function so that it is only imported if requested
"""


def get_alphatwirl():
    from . import alphatwirl
    return alphatwirl


def get_coffea():
    from . import coffea
    return coffea


known_backends = {"multiprocessing": get_alphatwirl,
                  "htcondor": get_alphatwirl,
                  "sge": get_alphatwirl,
                  "alphatwirl:multiprocessing": get_alphatwirl,
                  "alphatwirl:htcondor": get_alphatwirl,
                  "alphatwirl:sge": get_alphatwirl,
                  "coffea:local": get_coffea,
                  }


def get_backend(name):
    if name not in known_backends:
        raise ValueError("Unknown backend requested, '%s'" % name)
    return known_backends[name]()



"""
Summarize the data by producing binned and possibly weighted counts of the data.
"""
import os
import re
import numpy as np
import pandas as pd
from pandas.api.types import is_object_dtype
from . import binning_config as cfg


class Collector():
    valid_ext = {'xlsx': 'excel', 'h5': 'hdf', 'msg': 'msgpack', 'dta': 'stata', 'pkl': 'pickle', 'p': 'pickle'}

    def __init__(self, filename, dataset_col, binnings, file_format):
        self.filename = filename
        self.dataset_col = dataset_col
        self.binnings = binnings
        self.file_format = file_format

    def collect(self, dataset_readers_list, doReturn=True, writeFiles=True):
        if len(dataset_readers_list) == 0:
            if doReturn:
                return pd.DataFrame()
            else:
                return None

        output = self._prepare_output(dataset_readers_list)

        if writeFiles:
            for file_dict in self.file_format:
                file_ext = file_dict.pop('extension', None)
                save_func = file_ext.split('.')[1]
                if save_func in Collector.valid_ext:
                    save_func = Collector.valid_ext[save_func]
                try:
                    getattr(output, "to_%s" % save_func)(self.filename + file_ext, **file_dict)
                except AttributeError as err:
                    print("Incorrect file format: %s" % err)
                except TypeError as err:
                    print("Incorrect args: %s" % err)

        if doReturn:
            return output

    def _prepare_output(self, dataset_readers_list):
        return combined_dataframes(dataset_readers_list,
                                   self.dataset_col,
                                   binnings=self.binnings)


def combined_dataframes(dataset_readers_list, dataset_col, binnings=None):
    dataset_readers_list = [(d, [r.contents for r in readers])
                            for d, readers in dataset_readers_list if readers]
    if not dataset_readers_list:
        return None

    if dataset_col:
        output = _merge_dataframes(dataset_readers_list)
    else:
        output = _sum_dataframes(dataset_readers_list)
    if binnings:
        output = densify_dataframe(output, binnings)
    return output


def _merge_dataframes(dataset_readers_list):
    all_dfs = []
    keys = []
    for dataset, readers in dataset_readers_list:
        dataset_df = readers[0]
        for df in readers[1:]:
            if df is None:
                continue
            dataset_df = dataset_df.add(df, fill_value=0.)
        all_dfs.append(dataset_df)
        keys.append(dataset)
    final_df = pd.concat(all_dfs, keys=keys, names=['dataset'], sort=True)
    return final_df


def _sum_dataframes(dataset_readers_list):
    final_df = None
    for _, readers in dataset_readers_list:
        for df in readers:
            if df is None:
                continue
            if final_df is None:
                final_df = df
                continue
            final_df = final_df.add(df, fill_value=0.)
    return final_df


def densify_dataframe(in_df, binnings):
    in_index = in_df.index
    index_values = []
    for dim in in_index.names:
        bins = binnings.get(dim, None)
        if bins is None:
            index_values.append(in_index.unique(dim))
            continue
        index_values.append(bins)
    out_index = pd.MultiIndex.from_product(index_values, names=in_index.names)
    out_df = in_df.reindex(index=out_index, copy=False)
    return out_df


class BinnedDataframe():
    """Produces a binned dataframe (a multi-dimensional histogram).

    def __init__(self, name, out_dir, binning, weights=None, dataset_col=False):

    Parameters:
      binning (list[dict]): A list of dictionaries describing the variables to
        bin on, and how they should be binned.  Each of these dictionaries can
        contain the following:

        +-----------+----------------+--------------------------------------------------------------+
        | Parameter | Default        | Description                                                  |
        +===========+================+==============================================================+
        | ``in``    |                | The name of the attribute on the event to use.               |
        +-----------+----------------+--------------------------------------------------------------+
        | ``out``   | same as ``in`` | The name of the column to be filled in the output dataframe. |
        +-----------+----------------+--------------------------------------------------------------+
        | ``bins``  | ``None``       | | Must be either ``None`` or a dictionary.  If a dictionary, |
        |           |                |   it must contain one of the follow sets of                  |
        |           |                | | key-value pairs:                                           |
        |           |                | |   1. ``nbins``, ``low``, ``high``: which are used to       |
        |           |                |     produce a list of bin edges equivalent to:               |
        |           |                | |      ``numpy.linspace(low, high, nbins + 1)``              |
        |           |                | |   2. ``edges``: which is treated as the list of bin        |
        |           |                |     edges directly.                                          |
        |           |                | | If set to ``None``, then the input variable is assumed     |
        |           |                |   to already be categorical (ie. binned or discrete)         |
        +-----------+----------------+--------------------------------------------------------------+

      weights (str or list[str], dict[str, str]): How to weight events in the
        output table.  Must be either a single variable, a list of
        variables, or a dictionary where the values are variables in the data and
        keys are the column names that these weights should be called in the
        output tables.
      file_format (str or list[str], dict[str, str]): determines the file format to
        use to save the binned dataframe to disk.  Should be either a) a string with
        the file format, b) a dict containing the keyword `extension` to give the file
        format and then all other keyword-argument pairs are passed on to the
        corresponding pandas function, or c) a list of values matching a) or b).
      dataset_col (bool): adds an extra binning column with the name for each dataset.
      pad_missing (bool): If ``False``, any bins that don't contain data are
        excluded from the stored dataframe.  Leaving this ``False`` can save
        some disk-space and improve processing time, particularly if the bins are
        only very sparsely filled.

    Other Parameters:
      name (str):  The name of this stage (handled automatically by fast-flow)
      out_dir (str):  Where to put the summary table (handled automatically by
          fast-flow)

    Raises:
      BadBinnedDataframeConfig: If there is an issue with the binning description.

    """

    def __init__(self, name, out_dir, binning, weights=None, dataset_col=True, pad_missing=False, file_format=None):
        self.name = name
        self.out_dir = out_dir
        ins, outs, binnings = cfg.create_binning_list(self.name, binning)
        self._bin_dims = ins
        self.potential_inputs = set(sum((re.findall(r"\w+", dim) for dim in self._bin_dims), []))
        self._out_bin_dims = outs
        self._binnings = binnings
        self._dataset_col = dataset_col
        self._weights = cfg.create_weights(self.name, weights)
        self._pad_missing = pad_missing
        self._file_format = cfg.create_file_format(self.name, file_format)
        self.contents = None

    def collector(self):
        outfilename = "tbl_"
        if self._dataset_col:
            outfilename += "dataset."
        outfilename += ".".join(self._out_bin_dims)
        outfilename += "--" + self.name
        outfilename = os.path.join(self.out_dir, outfilename)
        binnings = None
        if self._pad_missing:
            binnings = dict(zip(self._out_bin_dims, self._binnings))
        return Collector(outfilename, self._dataset_col, binnings=binnings, file_format=self._file_format)

    def event(self, chunk):
        all_inputs = [key for key in chunk.tree.keys() if key.decode() in self.potential_inputs]
        if chunk.config.dataset.eventtype == "mc":
            weights = list(self._weights.values())
            all_inputs += weights
        else:
            weights = None

        data = chunk.tree.pandas.df(all_inputs, flatten=False)
        data = explode(data)

        binned_values = _bin_values(data, dimensions=self._bin_dims,
                                    binnings=self._binnings,
                                    weights=weights,
                                    out_weights=self._weights.keys(),
                                    out_dimensions=self._out_bin_dims)
        if self.contents is None:
            self.contents = binned_values
        else:
            self.contents = self.contents.add(binned_values, fill_value=0)

        return True

    def merge(self, rhs):
        if rhs.contents is None or len(rhs.contents) == 0:
            return
        if self.contents is None:
            self.contents = rhs.contents
            return
        self.contents = self.contents.add(rhs.contents, fill_value=0)


count_label = "n"
weight_labels = ("sumw", "sumw2")


def _make_column_labels(weights):
    labels = [w + ":" + l for l in weight_labels for w in weights]
    return [count_label] + labels


def _bin_values(data, dimensions, binnings, weights, out_dimensions=None, out_weights=None):
    if not out_dimensions:
        out_dimensions = dimensions
    if not out_weights:
        out_weights = weights

    final_bin_dims = []
    for dimension, binning in zip(dimensions, binnings):
        if binning is None:
            final_bin_dims.append(dimension)
            continue
        out_dimension = dimension + "_bins"
        data[out_dimension] = pd.cut(data.eval(dimension, engine='numexpr'), binning, right=False)
        final_bin_dims.append(out_dimension)

    if weights:
        weight_sq_dims = [w + "_squared" for w in weights]
        data[weight_sq_dims] = data[weights] ** 2

    bins = data.groupby(final_bin_dims)
    counts = bins[data.columns[0]].count()

    if weights:
        sums = bins[weights].sum()
        sum_sqs = bins[weight_sq_dims].sum()
        histogram = pd.concat([counts, sums, sum_sqs], axis="columns")
        histogram.columns = _make_column_labels(out_weights)
    else:
        histogram = counts.to_frame(count_label)

    histogram.index.set_names(out_dimensions, inplace=True)
    return histogram


_explodable_types = (tuple, list, np.ndarray)


def explode(df):
    """
    Based on this answer:
    https://stackoverflow.com/questions/12680754/split-explode-pandas\
    -dataframe-string-entry-to-separate-rows/40449726#40449726
    """
    # get the list columns
    lst_cols = [col for col, dtype in df.dtypes.items() if is_object_dtype(dtype)]
    # Be more specific about which objects are ok
    lst_cols = [col for col in lst_cols if isinstance(df[col][0], _explodable_types)]
    if not lst_cols:
        return df

    # all columns except `lst_cols`
    idx_cols = df.columns.difference(lst_cols)

    # check all lists have same length
    lens = pd.DataFrame({col: df[col].str.len() for col in lst_cols})
    different_length = (lens.nunique(axis=1) > 1).any()
    if different_length:
        raise ValueError("Cannot bin multiple arrays with different jaggedness")
    lens = lens[lst_cols[0]]

    # create "exploded" DF
    flattened = {col: df.loc[lens > 0, col].values for col in lst_cols}
    flattened = {col: sum(map(list, vals), []) for col, vals in flattened.items()}
    res = pd.DataFrame({col: np.repeat(df[col].values, lens) for col in idx_cols})
    res = res.assign(**flattened)

    # Check that rows are fully "exploded"
    return explode(res)



import inspect
import sys
import re
from . import known_stages


class StageGuidanceHelper:
    _common_config = ["name", "out_dir", "self"]

    def __init__(self, stage_class, module_name):
        self._stage = stage_class
        self.module_name = module_name

    @property
    def stage(self):
        return self._stage

    @property
    def class_name(self):
        return self.module_name + "." + self.stage.__name__

    def matches(self, regex):
        if regex:
            return re.search(regex, self.class_name)
        return True

    def parameters(self):
        args, vargs, kwargs, defaults, _, _, annots = get_signature(self.stage.__init__)
        args = [a for a in args if a not in self._common_config]
        return format_signature(args, vargs, kwargs, defaults, annots)

    def docstring(self, nlines=-1):
        doc = inspect.getdoc(self.stage)
        if not doc:
            return "<Missing docstring>"
        if nlines > 0:
            output = []
            while nlines > 0 and doc:
                lines = doc.split("\n", nlines)
                lines, remainder = lines[:nlines], lines[nlines:]
                lines = [l for l in lines if l.strip()]
                output += lines
                if not remainder:
                    break
                doc = remainder[0].strip()
                nlines -= len(lines)
            doc = "\n".join(output)
        return doc


def get_signature(function):
    if sys.version[0] < "3":
        return inspect.getargspec(function) + (None, None, None)
    return inspect.getfullargspec(function)


def format_signature(args, vargs, kwargs, defaults, annots):
    if annots:
        args = ["%s:%s" % (a, annots[a].__name__) if a in annots else a for a in args]
    if defaults:
        ndefs = min(len(defaults), len(args))
        def_args = ["{}={}".format(a, d) for a, d in zip(args[-ndefs:], defaults[-ndefs:])]
        args = args[:-ndefs] + def_args
    if vargs:
        args.append("*" + vargs)
    if kwargs:
        args.append("**" + kwargs)
    return args


all_stages = tuple(StageGuidanceHelper(s, "fast_carpenter") for s in known_stages)


def help_stages(stage_name, full_output=False):
    stages = all_stages
    if stage_name and stage_name.lower() != "all":
        stages = tuple(s for s in stages if s.matches(stage_name))
        if not stages:
            raise RuntimeError("Unknown stage:", stage_name)

    for i, stage in enumerate(stages):
        name = stage.class_name
        args = ", ".join(stage.parameters())
        header = "{name}\n   config: {args}".format(name=name, args=args)
        print(header)
        if full_output:
            print(stage.docstring())
        else:
            print("   purpose:", stage.docstring(1))
        if i != len(stages) - 1:
            end_stage = "=" * 80 + "\n" if full_output else ""
            print(end_stage)



from __future__ import absolute_import
import os
from collections import defaultdict
import numpy as np
from . import binning_config as cfg
from . import binned_dataframe as binned_df
from .import_aghast import aghast


class Collector():
    def __init__(self, filename, axes, edges, by_dataset):
        self.filename = filename
        self.axes = axes
        self.edges = edges
        self.by_dataset = by_dataset

    def collect(self, dataset_readers_list):
        if len(dataset_readers_list) == 0:
            return None

        dataframe = binned_df.combined_dataframes(dataset_readers_list,
                                                  self.by_dataset,
                                                  binnings=self.edges)
        full_axes = complete_axes(self.axes, dataframe.index)
        dataframe.fillna(0, inplace=True)
        dataframe.to_csv(self.filename + ".csv")
        all_counters = convert_to_counters(dataframe)
        collection = aghast.Collection(all_counters, full_axes)
        collection.tofile(self.filename + ".ghst")


def convert_to_counters(df):
    counts = aghast.UnweightedCounts(aghast.InterpretedInlineFloat64Buffer(df[binned_df.count_label].values))
    counters = {binned_df.count_label: aghast.Histogram([aghast.Axis()], counts)}
    weight_labels = defaultdict(dict)
    for col in df.columns:
        if col == binned_df.count_label:
            continue
        label, sumtype = col.split(":")
        weight_labels[label][sumtype] = col
    for label, sums in weight_labels.items():
        sumw_col = sums[binned_df.weight_labels[0]]
        sumw = aghast.InterpretedInlineFloat64Buffer(df[sumw_col].values.astype(np.float64))

        sumw2 = None
        sumw2_col = sums.get(binned_df.weight_labels[1], None)
        if sumw2_col:
            sumw2 = aghast.InterpretedInlineFloat64Buffer(df[sumw2_col].values.astype(np.float64))
        counters[label] = aghast.Histogram([aghast.Axis()], aghast.WeightedCounts(sumw, sumw2))
    return counters


def complete_axes(axes, df_index):
    full_axes = []
    for dim in df_index.names:
        ax = axes.get(dim, None)
        if ax is None:
            values = df_index.unique(dim).astype(str)
            full_axes.append(aghast.Axis(aghast.CategoryBinning(values), title=dim))
        else:
            full_axes.append(ax)
    return full_axes


def _ovf_convention():
    return aghast.RealOverflow(loc_underflow=aghast.BinLocation.below1,
                               loc_overflow=aghast.BinLocation.above1)


def bin_one_dimension(low=None, high=None, nbins=None, edges=None,
                      **kwargs):
    # - bins: {nbins: 6 , low: 1  , high: 5 , overflow: True}
    # - bins: {edges: [0, 200., 900], overflow: True}
    if all([x is not None for x in (nbins, low, high)]):
        aghast_bins = aghast.RegularBinning(nbins,
                                            aghast.RealInterval(low, high),
                                            overflow=_ovf_convention())
    elif edges:
        # array are fixed to float type, to be consistent with the float-type underflow and overflow bins
        edges = np.array(edges, "f")
        aghast_bins = aghast.EdgesBinning(edges, overflow=_ovf_convention())
    else:
        return None
    return aghast_bins


class BuildAghast:
    """Builds an aghast histogram.

    Can be parametrized in the same way as
    :py:class:`fast_carpenter.BinnedDataframe` (and actually uses that stage
    behind the scenes) but additionally writes out a Ghast which can be
    reloaded with other ghast packages.

    .. seealso::

       * :py:class:`fast_carpenter.BinnedDataframe` for a version which only
         produces binned pandas dataframes.
       * The aghast main page: `<https://github.com/scikit-hep/aghast>`_.

    """

    def __init__(self, name, out_dir, binning, weights=None, dataset_col=True):
        self.name = name
        self.out_dir = out_dir
        self.builder = binned_df.BinnedDataframe(name + "_builder", out_dir,
                                                 binning=binning,
                                                 weights=weights,
                                                 dataset_col=dataset_col)
        ins, outs, bins = cfg.create_binning_list(self.name, binning, make_bins=bin_one_dimension)
        self.axes = {_out: aghast.Axis(_bins, title=_out, expression=_in) if _bins else None
                     for _in, _out, _bins in zip(ins, outs, bins)}
        self.by_dataset = dataset_col

    @property
    def contents(self):
        return self.builder.contents

    def collector(self):
        outfilename = "tbl_"
        if self.by_dataset:
            outfilename += "dataset."
        outfilename += ".".join(self.axes.keys())
        outfilename += "--" + self.name
        outfilename = os.path.join(self.out_dir, outfilename)
        edges = dict(zip(self.builder._out_bin_dims, self.builder._binnings))
        return Collector(outfilename, axes=self.axes,
                         edges=edges, by_dataset=self.by_dataset)

    def event(self, chunk):
        return self.builder.event(chunk)

    def merge(self, rhs):
        self.builder.merge(rhs.builder)



from .binned_dataframe import BinnedDataframe
from .event_level_dataframe import EventByEventDataframe
from .aghast import BuildAghast

__all__ = ["BuildAghast", "BinnedDataframe", "EventByEventDataframe"]



from .variables import Define
import six


class BadSystematicWeightsConfig(Exception):
    pass


class SystematicWeights():
    """Combines multiple weights and variations to produce a single event weight

    To study the impact of systematic uncertainties it is common to re-weight events
    using a variation of the weights representing, for example, a 1-sigma increase or decrease
    in the weights.  Once there are multiple weight schemes involved writing out each possible combination
    of these weights becomes tedious and potentially error-prone; this stage makes it easier.

    It forms the ``nominal`` weight for each event by multiplying all nominal weights together,
    then the specific variation by replacing a given nominal weight with its corresponding "up" or "down" variation.

    Each variation of a weight should just be a string giving an expression to
    use for that variation.  This stage then combines these into a single
    expression by joining each set of variations with `"*"`, i.e. multiplying
    them together.  The final results then use an internal :py:class:`Define`
    stage to do the calculation.

    Parameters:
      weights (dictionary[str, dictionary]):  A Dictionary of weight variations
          to combine.  The keys in this dictionary will determine how this
          variation is called in the output variable.  The values of this
          dictionary should either be a single string -- the name of the input
          variable to use for the "nominal" variation, or a dictionary containing
          any of the keys, ``nominal``, ``up``, or ``down``.  Each of these should
          then have a value providing the expression to use for that variation/
      out_format (str): The format string to use to build the name of the
          output variations.  Defaults to "weight_{}".  Should contain a pair
          of empty braces which will be replaced with the name for the current
          variation, e.g. "nominal" or "PileUp_up".
      extra_variations (list[str]): A list of additional variations to allow

    Other Parameters:
      name (str):  The name of this stage (handled automatically by fast-flow)
      out_dir (str):  Where to put the summary table (handled automatically by
          fast-flow)

    Example:
      ::

        syst_weights:
          energy_scale: {nominal: WeightEnergyScale, up: WeightEnergyScaleUp, down: WeightEnergyScaleDown}
          trigger: TriggerEfficiency
          recon: {nominal: ReconEfficiency, up: ReconEfficiency_up}

      which will create 4 new variables:
      ::

        weight_nominal =  WeightEnergyScale * TriggerEfficiency * ReconEfficiency
        weight_energy_scale_up =  WeightEnergyScaleUp * TriggerEfficiency * ReconEfficiency
        weight_energy_scale_down =  WeightEnergyScaleDown * TriggerEfficiency * ReconEfficiency
        weight_recon_up =  WeightEnergyScale * TriggerEfficiency * ReconEfficiency_up
    """
    def __init__(self, name, out_dir, weights, out_format="weight_{}", extra_variations=[]):
        self.name = name
        self.out_dir = out_dir
        weights = _normalize_weights(name, weights, tuple(extra_variations))
        variations = _build_variations(name, weights, out_fmt=out_format)
        self.variable_maker = Define(name + "_builder", out_dir, variations)

    def event(self, chunk):
        if not chunk.config.dataset.eventtype == "mc":
            return True
        return self.variable_maker.event(chunk)


def _normalize_weights(stage_name, variable_list, valid_vars):
    if not isinstance(variable_list, dict):
        msg = "{}: Didn't receive a list of variables"
        raise BadSystematicWeightsConfig(msg.format(stage_name))
    return {name: _normalize_one_variation(stage_name, cfg, name, valid_vars=valid_vars)
            for name, cfg in variable_list.items()}


def _build_variations(stage_name, weights, out_fmt="weight_{}"):
    def _combine_weights(w):
        return "(" + ")*(".join(w) + ")"

    nominal_weights = {n: w["nominal"] for n, w in weights.items()}
    variations = [{out_fmt.format("nominal"): _combine_weights(nominal_weights.values())}]
    weights_to_vary = {(n, var): w[var] for n, w in weights.items() for var in w if var != "nominal"}
    for (name, direction), variable in weights_to_vary.items():
        combination = nominal_weights.copy()
        combination[name] = variable
        combination = _combine_weights(combination.values())
        variations.append({out_fmt.format(name + "_" + direction): combination})
    return variations


def _normalize_one_variation(stage_name, cfg, name, valid_vars):
    if isinstance(cfg, six.string_types):
        return dict(nominal=cfg)
    if not isinstance(cfg, dict):
        msg = "{}: Each systematic weight should be either a dict or just a string"
        raise BadSystematicWeightsConfig(msg.format(stage_name))
    if "nominal" not in cfg:
        msg = "{}: No nominal weight provided for '{}'"
        raise BadSystematicWeightsConfig(msg.format(stage_name, name))

    bad_keys = [key for key in cfg if key not in ("nominal", "up", "down") + valid_vars]
    if bad_keys:
        msg = "{}: Received unknown keys,'{}', for '{}'"
        raise BadSystematicWeightsConfig(msg.format(stage_name, bad_keys, name))
    return cfg



"""
Functions to run a job using alphatwirl
"""


class DummyCollector():
    def collect(self, *args, **kwargs):
        pass


class AtuprootContext:
    def __enter__(self):
        import atuproot.atuproot_main as atup
        self.atup = atup
        self._orig_event_builder = atup.EventBuilder
        self._orig_build_parallel = atup.build_parallel

        from ..event_builder import EventBuilder
        from atsge.build_parallel import build_parallel
        atup.EventBuilder = EventBuilder
        atup.build_parallel = build_parallel
        return self

    def __exit__(self, *args, **kwargs):
        self.atup.EventBuilder = self._orig_event_builder
        self.atup.build_parallel = self._orig_build_parallel


def execute(sequence, datasets, args):
    """
    Run a job using alphatwirl and atuproot
    """

    if args.ncores < 1:
        args.ncores = 1

    sequence = [(s, s.collector() if hasattr(s, "collector") else DummyCollector()) for s in sequence]

    with AtuprootContext() as runner:
        process = runner.atup.AtUproot(args.outdir,
                                       quiet=args.quiet,
                                       parallel_mode=args.mode,
                                       process=args.ncores,
                                       max_blocks_per_dataset=args.nblocks_per_dataset,
                                       max_blocks_per_process=args.nblocks_per_sample,
                                       nevents_per_block=args.blocksize,
                                       profile=args.profile,
                                       profile_out_path="profile.txt",
                                       )

        ret_val = process.run(datasets, sequence)

    if not args.profile:
        # This breaks in AlphaTwirl when used with the profile option
        summary = {s[0].name: list(df.index.names) for s, df in zip(sequence, ret_val[0]) if df is not None}
    else:
        summary = " (Results summary not available with profile mode) "

    return summary, ret_val



"""
Functions to run a job using Coffea
"""
import copy
from fast_carpenter.masked_tree import MaskedUprootTree
from collections import namedtuple
from coffea import processor
from coffea.processor import futures_executor, run_uproot_job


EventRanger = namedtuple("EventRanger", "start_entry stop_entry entries_in_block")
SingleChunk = namedtuple("SingleChunk", "tree config")
ChunkConfig = namedtuple("ChunkConfig", "dataset")
ConfigProxy = namedtuple("ConfigProxy", "name eventtype")


class stages_accumulator(processor.AccumulatorABC):
    def __init__(self, stages):
        self._zero = copy.deepcopy(stages)
        self._value = copy.deepcopy(stages)

    def identity(self):
        return stages_accumulator(self._zero)

    def __getitem__(self, idx):
        return self._value[idx]

    def add(self, other):
        for i, stage in enumerate(self._value):
            if not hasattr(stage, "merge"):
                continue
            stage.merge(other[i])


class FASTProcessor(processor.ProcessorABC):
    def __init__(self, sequence):

        self._columns = list()
        self._sequence = sequence
        accumulator_dict = {'stages': processor.dict_accumulator({})}
        self._accumulator = processor.dict_accumulator(accumulator_dict)

    @property
    def columns(self):
        return self._columns

    @property
    def accumulator(self):
        return self._accumulator

    def process(self, df):
        output = self.accumulator.identity()

        start = df._branchargs['entrystart']
        stop = df._branchargs['entrystop']
        tree = MaskedUprootTree(df._tree, EventRanger(start, stop, stop - start))
        dsname = df['dataset']
        cfg_proxy = ConfigProxy(dsname, 'data' if dsname == 'data' else 'mc')
        chunk = SingleChunk(tree, ChunkConfig(cfg_proxy))

        output['stages'][dsname] = stages_accumulator(self._sequence)

        for work in output['stages'][dsname]._value:
            work.event(chunk)

        return output

    def postprocess(self, accumulator):
        stages = accumulator['stages']
        results = {}

        wf = copy.deepcopy(self._sequence)
        for i_step, step in enumerate(wf):
            if not hasattr(step, "collector"):
                continue
            collector = step.collector()
            output = collector.collect([(d, (s[i_step],)) for d, s in stages.items()])
            results[step.name] = output

        accumulator['results'] = results

        return accumulator


def execute(sequence, datasets, args):
    fp = FASTProcessor(sequence)

    executor = futures_executor
    exe_args = {'workers': args.ncores,
                'chunksize': args.blocksize,
                'maxchunks': args.nblocks_per_dataset,
                'function_args': {'flatten': False}}

    coffea_datasets = {}
    for ds in datasets:
        coffea_datasets[ds.name] = ds.__dict__.copy()
        coffea_datasets[ds.name].pop('name')
        coffea_datasets[ds.name]['treename'] = coffea_datasets[ds.name].pop('tree')

    out = run_uproot_job(coffea_datasets, 'events', fp, executor, executor_args=exe_args)
    return out["stages"], out["results"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

import os
from setuptools import setup, find_packages


with open('README.rst') as readme_file:
    readme = readme_file.read()

# with open('HISTORY.rst') as history_file:
#     history = history_file.read()


def get_version():
    _globals = {}
    with open(os.path.join("fast_carpenter", "version.py")) as version_file:
        exec(version_file.read(), _globals)
    return _globals["__version__"]


requirements = ['atuproot==0.1.13', 'atsge==0.2.1', 'fast-flow', 'fast-curator', 'awkward',
                'pandas', 'numpy', 'numba', 'numexpr', 'uproot>=3']
repositories = []

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest', 'flake8', 'pytest-cov']

setup(
    author="Ben Krikler",
    author_email='fast-hep@cern.ch',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    description="F.A.S.T. package for summarizing ROOT TTrees",
    entry_points={
        'console_scripts': [
            'fast_carpenter=fast_carpenter.__main__:main',
        ],
    },
    install_requires=requirements,
    dependency_links=repositories,
    license="Apache Software License 2.0",
    long_description=readme,  # + '\n\n' + history,
    include_package_data=True,
    keywords=['ROOT', 'pandas', 'analysis', 'particle physics', 'HEP', 'F.A.S.T'],
    name='fast-carpenter',
    packages=find_packages(include=['fast_carpenter*']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/FAST-HEP/fast-carpenter',
    version=get_version(),
    zip_safe=True,
)



# -*- coding: utf-8 -*-

"""Top-level package for fast-carpenter."""

__author__ = """Benjamin Krikler, and F.A.S.T"""
__email__ = 'fast-hep@cern.ch'


from .define.variables import Define
from .define.systematics import SystematicWeights
from .selection.stage import CutFlow, SelectPhaseSpace
from .summary import BinnedDataframe, BuildAghast, EventByEventDataframe
from .version import __version__, version_info


__all__ = ["Define", "SystematicWeights", "CutFlow",
           "SelectPhaseSpace", "BinnedDataframe", "BuildAghast",
           "__version__", "version_info"]


known_stages = [Define, SystematicWeights, CutFlow,
                SelectPhaseSpace, BinnedDataframe, BuildAghast,
                EventByEventDataframe]
